/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "recovery.h"

static unsigned char	csumbuff[ BUFFER_SIZE_BYTE ];

static inline unsigned long
checksum_buff( unsigned long sum, unsigned char * buf, int len )
{
	int		i, n;
	unsigned long	s, h, l;

	s = (~ntohl(sum)) & 0xFFFF;

	for( i = 0, n = len - 2; i < n; i += 2 )
	{
		h = *buf++;
		l = *buf++;
		s += ( ( h << 8 ) | l );
        }

	if ( len & 1 )
	{
		h = *buf++;
		l = 0;
	}
	else
	{
		h = *buf++;
		l = *buf++;
	}

	s += ( ( h << 8 ) | l );
	s = ( s & 0xFFFF ) + ( s >> 16 );
	sum = htonl( ( ~( ( s & 0xFFFF ) + ( s >> 16 ) ) ) & 0xFFFF );

	return sum;
}

static inline unsigned long
checksum_reg( char * file, unsigned long off, unsigned long len )
{
	unsigned long   sum = 0;
	int		fd;
	unsigned long	blen = BUFFER_SIZE_BYTE, sumlen = 0, rlen;

	if ( ( fd = open( file, O_RDONLY ) ) == -1 )
		goto out;
	if ( lseek( fd, off, SEEK_SET ) == -1 )
		goto out_close;
	blen = ( ( blen < len ) ? blen : len );

	while ( sumlen < len &&
		( rlen = read( fd, csumbuff, blen ) ) > 0 )
	{
		sum = checksum_buff( sum, csumbuff, rlen );
		sumlen += rlen;
	}

out_close:
	close( fd );
out:
	return	sum;
}

static inline unsigned long
checksum_dir( char * file )
{
	struct dirent *	ent;
	DIR *		dir;
	unsigned long   sum = 0;

	if ( !( dir = opendir( file ) ) )
		goto out;
	while ( (  ent = readdir( dir ) ) )
		sum = checksum_buff( sum,
				ent->d_name, strlen( ent->d_name ) );
	closedir( dir );
out:
	return	sum;
}

static inline unsigned long
checksum_lnk( char * file )
{
	unsigned long	sum = 0;
	int		len;

	if ( ( len = readlink( file, csumbuff, BUFFER_SIZE_BYTE ) ) != -1 )
		sum = checksum_buff( sum, csumbuff, len );

        return	sum;
}

unsigned long
checksum( unsigned long type, char * file,
		unsigned long off, unsigned long len )
{
	if ( file )
	{
		switch ( type )
		{
		case TYPE_REG:
			return	checksum_reg( file, off, len );

		case TYPE_DIR:
			return	checksum_dir( file );

		case TYPE_LNK:
			return	checksum_lnk( file );
		}
	}
	return	0;
}
