/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "recovery.h"

char	pid_file[1024]		= "/var/run/recovery.pid";
char	std_logfile[1024]	= "/var/log/recovery_log";
char	err_logfile[1024]	= "/var/log/recovery_err";

FILE *	stdFp;
FILE *	errFp;

key_t	recovery_key	= (key_t)2048;
short	check_port	= 2048;
short	repair_port	= 2049;
short	pmap_port	= 111;

#if 0
#define	PID_FILE	"/var/run/recovery.pid"
#define	STD_LOGFILE	"/var/log/recovery_log"
#define	ERR_LOGFILE	"/var/log/recovery_err"
#endif

#if 0
void
openLog()
{
	stdFp = stdout;
	errFp = stderr;
}

void
closeLog()
{

}
#else
void
openLog()
{
	stdFp = fopen( std_logfile, "a" );
	if ( !stdFp )	stdFp = stdout;
	errFp = fopen( err_logfile, "a" );
	if ( !errFp )	errFp = stderr;
}

void
closeLog()
{
	if ( stdFp && stdFp != stdout )
	{
		fclose( stdFp );
		stdFp = NULL;
	}
	if ( errFp && errFp != stderr )
	{
		fclose( errFp );
		errFp = NULL;
	}
}
#endif

volatile sig_atomic_t  sig_num;

void
sig_handler( int sig_no )
{
        if ( !sig_num )
                sig_num = sig_no;
}

static inline int
recovery( pid_t c_pid, pid_t r_pid )
{
	pid_t	pid = fork();
	struct sigaction act;
	int	qid;
	time_t	t;

	if ( pid ) return pid;

	setpgid( 0, 0 );

	act.sa_handler = sig_handler;
	sigemptyset( &act.sa_mask );
	act.sa_flags = 0;
	sigaction( SIGHUP, &act, 0 );
	sigaction( SIGINT, &act, 0 );

	openLog();
	t = time( NULL );
	fprintf( stdFp, "recovery boot : %s", ctime( &t ) );
	do
	{
		sig_num = 0;
#if 0
		if ( ( qid = msgget( IPC_DEFAULT_KEY,
				0666 | IPC_CREAT ) ) == -1 )
		{
			fprintf( errFp, "msgget failed!\n" );
			goto out;
		}
#else
		if ( ( qid = msgget( recovery_key,
				0666 | IPC_CREAT ) ) == -1 )
		{
			fprintf( errFp, "msgget failed!\n" );
			goto out;
		}
#endif
#if 1
		init_pool( qid );
#endif
		recovery_main( qid );
		kill( c_pid, sig_num );
		kill( r_pid, sig_num );

		if ( sig_num == SIGHUP )
		{
			t = time( NULL );
			fprintf( stdFp, "recovery reboot : %s", ctime( &t ) );
		}
		else
		{
			goto out;
		}
	}
	while( 1 );
out:
	t = time( NULL );
	fprintf( stdFp, "recovery terminate : %s", ctime( &t ) );

	closeLog();

	exit( 0 );
}

static inline pid_t
check()
{
	struct sockaddr_in	sock;
	struct sigaction	act;
	pid_t			pid = fork();
	int			sock_fd;
	time_t			t;

	if ( pid ) return pid;

	setpgid( 0, 0 );

	openLog();
	t = time( NULL );

	fprintf( stdFp, "check boot : %s", ctime( &t ) );

	act.sa_handler = sig_handler;
	sigemptyset( &act.sa_mask );
	act.sa_flags = 0;
	sigaction( SIGHUP, &act, 0 );
	sigaction( SIGINT, &act, 0 );

	do
	{
		sock.sin_family = PF_INET;
		sock.sin_addr.s_addr = htonl( INADDR_ANY );
#if 0
		sock.sin_port = TARGET_CHECK_PORT;
#else
		sock.sin_port = htons( check_port );
#endif

		if ( ( sock_fd = tcp_bind( &sock ) ) != -1 )
		{
			sig_num = 0;
			check_main( sock_fd );

			if ( sig_num == SIGINT )
				goto out;

			t = time( NULL );
			fprintf( stdFp, "check reboot : %s", ctime( &t ) );
			close( sock_fd );
		}
		else
		{
			sleep( 5 );
		}
	}
	while( 1 );
out:
	t = time( NULL );
	fprintf( stdFp, "check terminate : %s", ctime( &t ) );
	closeLog();

	exit( 0 );
}

static inline pid_t
repair()
{
	struct sockaddr_in	sock;
	struct sigaction	act;
	pid_t			pid = fork();
	int			sock_fd;
	time_t			t;

	if ( pid ) return pid;

	setpgid( 0, 0 );

	openLog();
	t = time( NULL );

	fprintf( stdFp, "repair boot : %s", ctime( &t ) );

	act.sa_handler = sig_handler;
	sigemptyset( &act.sa_mask );
	act.sa_flags = 0;
	sigaction( SIGHUP, &act, 0 );
	sigaction( SIGINT, &act, 0 );

	do
	{
		sock.sin_family = PF_INET;
		sock.sin_addr.s_addr = htonl( INADDR_ANY );
#if 0
		sock.sin_port = TARGET_REPAIR_PORT;
#else
		sock.sin_port = htons( repair_port );
#endif

		if ( ( sock_fd = tcp_bind( &sock ) ) != -1 )
		{
			sig_num = 0;
			repair_main( sock_fd );

			if ( sig_num == SIGINT )
				goto out;

			t = time( NULL );
			fprintf( stdFp, "repair reboot : %s", ctime( &t ) );
			close( sock_fd );
		}
		else
		{
			sleep( 5 );
		}
	}
	while( 1 );
out:
	t = time( NULL );
	fprintf( stdFp, "repair terminate : %s", ctime( &t ) );
	closeLog();

	exit( 0 );
}

#define CONF_REC_STD_LOG	"REC_STD_LOG"
#define CONF_REC_ERR_LOG	"REC_ERR_LOG"
#define CONF_REC_PID_FIL	"REC_PID_FIL"

#define CONF_REC_IPC_KEY	"REC_IPC_KEY"
#define CONF_REC_CHK_PORT	"REC_CHK_PORT"
#define CONF_REC_REP_PORT	"REC_REP_PORT"
#define CONF_REC_AUT_PORT	"REC_AUT_PORT"

static inline void
setConfig()
{
	char * value;

	if ( ( value = getenv( CONF_REC_STD_LOG ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %s > %s\n", CONF_REC_STD_LOG, std_logfile, value );
#endif
		strcpy( std_logfile, value );
#if 0
	}
printf( "REC_STD_LOG : %s\n", std_logfile );
#endif

	if ( ( value = getenv( CONF_REC_ERR_LOG ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %s > %s\n", CONF_REC_ERR_LOG, err_logfile, value );
#endif
		strcpy( err_logfile, value );
#if 0
	}
printf( "REC_ERR_LOG : %s\n", err_logfile );
#endif

	if ( ( value = getenv( CONF_REC_PID_FIL ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %s > %s\n", CONF_REC_PID_FIL, pid_file, value );
#endif
		strcpy( pid_file, value );
#if 0
	}
printf( "REC_PID_FIL : %s\n", pid_file );
#endif

	if ( ( value = getenv( CONF_REC_IPC_KEY ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %d > %d\n", CONF_REC_IPC_KEY,
				recovery_key, atoi( value ) );
#endif
		recovery_key = ( key_t )atoi( value );
#if 0
	}
printf( "REC_IPC_KEY : %d\n", recovery_key );
#endif

	if ( ( value = getenv( CONF_REC_CHK_PORT ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %d > %d\n", CONF_REC_CHK_PORT,
				check_port, atoi( value ) );
#endif
		check_port = ( unsigned short )atoi( value );
#if 0
	}
printf( "REC_CHK_PORT : %d\n", check_port );
#endif

	if ( ( value = getenv( CONF_REC_REP_PORT ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %d > %d\n", CONF_REC_REP_PORT,
				repair_port, atoi( value ) );
#endif
		repair_port = ( unsigned short )atoi( value );
#if 0
	}
printf( "REC_REP_PORT : %d\n", repair_port );
#endif

	if ( ( value = getenv( CONF_REC_AUT_PORT ) ) )
#if 0
	{
printf( "CONFIG CHANGE %s %d > %d\n", CONF_REC_AUT_PORT,
				pmap_port, atoi( value ) );
#endif
		pmap_port = ( unsigned short )atoi( value );
#if 0
	}
printf( "REC_AUT_PORT : %d\n", pmap_port );
#endif
}


int
main()
{
	FILE *	fp;
	pid_t   m_pid, c_pid, r_pid;

	if ( getuid() || getgid() ) 
		goto out;

	setConfig();

	if ( ( c_pid = check() ) < 0 )
		goto out;

	if ( ( r_pid = repair() ) < 0 )
		goto term_c_pid;

	if ( ( m_pid = recovery( c_pid, r_pid ) ) < 0 )
		goto term_r_pid;

	if ( !( fp = fopen( pid_file, "w" ) ) )
		fp = stdout;

	fprintf( fp, "%d", m_pid );
	if ( fp != stdout ) fclose( fp );
out:
	exit( 0 );
term_r_pid:
	kill( r_pid, SIGKILL );
term_c_pid:
	kill( c_pid, SIGKILL );
	goto out;
}
