/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "recovery.h"

static inline int
repair_request( int fd, buffer_t * buff,
		unsigned long rt, unsigned long ss, unsigned char * sp )
{
	unsigned long	status;

	buff->offset = buff->length = 0;
	if ( !free_u32length( buff, 3 ) )
		goto out_err;
	set_hlong( buff, COMMAND_REPAIR );
	set_hlong( buff, rt );
	set_nlong( buff, ss );
	if ( !set_hstr( buff, sp ) )
		goto out_err;
	if ( send_buffer( fd, buff ) < 0 )
		goto out_err;

	if ( receive_status( fd, &status ) < 0 )
		goto out_err;

	return status;
out_err:
	return -1;
}

static inline int
repair_type( recovery_t * rec, int i )
{
#if 0
fprintf( stdFp, "REPAIR TYPE REQUEST %s %s( %d ) ",
			rec->targ_pt,
			inet_ntoa( rec->serv_rp[ i ].serv_ip ),
			rec->serv_rp[ i ].mark );
fprintf( stdFp, "SUPPORT SERVER %s( %d )\n",
			inet_ntoa( rec->serv_rp[ rec->serv_ok ].serv_ip ),
			rec->serv_rp[ rec->serv_ok ].mark );
#endif
	return repair_request(	rec->serv_rp[ i ].serv_fd,
				&rec->sendbuff,
				rec->serv_rp[ i ].mark,
				rec->serv_rp[ rec->serv_ok ].serv_ip,
				rec->serv_rp[ rec->serv_ok ].serv_ex );
}

static inline int
majority_type_one( recovery_t * rec, int index )
{
	int	i = 0, c = 0;

	while ( i < rec->serv_ln )
	{
		if( rec->serv_rp[ index ].type == rec->serv_rp[ i ].type )
			c++;
		i++;
	}
	return c;
}

static inline int
majority_type( recovery_t * rec )
{
	int	i = 0, h = ( ( rec->serv_ln + 2 ) >> 1 ), rc = 0, c;

	rec->serv_ok = -1;
	while ( i < rec->serv_ln )
	{
		rec->serv_rp[ i ].mark = 0;
		if ( rec->serv_rp[ i ].serv_at == STATUS_OK &&
			( c = majority_type_one( rec, i ) ) < h )
		{
			rec->serv_rp[ i ].mark = REC_CREATE;
			rc++;
		}
		else if ( rec->serv_ok == -1 )
		{
			rec->serv_ok = i;
		}
		i++;
	}
	if ( rc == 0 || rec->serv_ok == -1 )	goto out;

	i = 0;
	while ( rc && i < rec->serv_ln )
	{
		if ( rec->serv_rp[ i ].mark == REC_CREATE )
		{
			if ( rec->serv_rp[ rec->serv_ok ].type == NO_EXISTS )
				rec->serv_rp[ i ].mark = REC_REMOVE;
			else if ( rec->serv_rp[ i ].type != NO_EXISTS )
				rec->serv_rp[ i ].mark |= REC_REMOVE;
			repair_type( rec, i );
		}
		i++;
	}
out:
	return 0;
}

static inline int
repair_attr( recovery_t * rec, int i )
{
#if 0
fprintf( stdFp, "REPAIR ATTR REQUEST %s %s( %d ) ",
			rec->targ_pt,
			inet_ntoa( rec->serv_rp[ i ].serv_ip ),
			rec->serv_rp[ i ].mark );
fprintf( stdFp, "SUPPORT SERVER %s( %d )\n",
			inet_ntoa( rec->serv_rp[ rec->serv_ok ].serv_ip ),
			rec->serv_rp[ rec->serv_ok ].mark );
#endif
	return repair_request(	rec->serv_rp[ i ].serv_fd,
				&rec->sendbuff,
				rec->serv_rp[ i ].mark,
				rec->serv_rp[ rec->serv_ok ].serv_ip,
				rec->serv_rp[ rec->serv_ok ].serv_ex );
}

static inline int
majority_attr_one( recovery_t * rec, int index )
{
	int	i = 0, c = 0;

	while ( i < rec->serv_ln )
	{
		if( rec->serv_rp[ index ].stat.st_mode ==
				rec->serv_rp[ i ].stat.st_mode &&
		    rec->serv_rp[ index ].stat.st_uid ==
				rec->serv_rp[ i ].stat.st_uid &&
		    rec->serv_rp[ index ].stat.st_gid ==
				rec->serv_rp[ i ].stat.st_gid &&
		    rec->serv_rp[ index ].stat.st_size ==
				rec->serv_rp[ i ].stat.st_size )
			c++;
		i++;
	}
	return c;
}

static inline int
majority_attr( recovery_t * rec )
{
	int	i = 0, h = ( ( rec->serv_ln + 2 ) >> 1 ), rc = 0;

	rec->serv_ok = -1;
	while ( i < rec->serv_ln )
	{
		if ( rec->serv_rp[ i ].serv_at == STATUS_OK &&
			!( rec->serv_rp[ i ].mark &
				( REC_CREATE | REC_REMOVE ) ) &&
			majority_attr_one( rec, i ) < h )
		{
			rec->serv_rp[ i ].mark = REC_ATTR;
			rc++;
		}
		else if ( rec->serv_ok == -1 )
		{
			rec->serv_ok = i;
		}
		i++;
	}
	if ( rc == 0 || rec->serv_ok == -1 )	goto out;

	i = 0;
	while ( rc && i < rec->serv_ln )
	{
		if ( rec->serv_rp[ i ].mark & REC_ATTR )
			repair_attr( rec, i );
		i++;
	}
out:
	return 0;
}

static inline int
repair_data( recovery_t * rec, int i )
{
#if 0
fprintf( stdFp, "REPAIR DATA REQUEST %s %s( %d ) ",
			rec->targ_pt,
			inet_ntoa( rec->serv_rp[ i ].serv_ip ),
			rec->serv_rp[ i ].mark );
fprintf( stdFp, "SUPPORT SERVER %s( %d )\n",
			inet_ntoa( rec->serv_rp[ rec->serv_ok ].serv_ip ),
			rec->serv_rp[ rec->serv_ok ].mark );
#endif
	return repair_request(	rec->serv_rp[ i ].serv_fd,
				&rec->sendbuff,
				rec->serv_rp[ i ].mark,
				rec->serv_rp[ rec->serv_ok ].serv_ip,
				rec->serv_rp[ rec->serv_ok ].serv_ex );
}

static inline int
majority_data_one( recovery_t * rec, int index )
{
	int	i = 0, c = 0;

	while ( i < rec->serv_ln )
	{
		if( rec->serv_rp[ index ].csum == rec->serv_rp[ i ].csum &&
		    rec->serv_rp[ index ].stat.st_size ==
					rec->serv_rp[ i ].stat.st_size )
			c++;
		i++;
	}
	return c;
}

static inline int
majority_data( recovery_t * rec )
{
	int	i = 0, h = ( ( rec->serv_ln + 2 ) >> 1 ), rc = 0;

	rec->serv_ok = -1;
	while ( i < rec->serv_ln )
	{
		if ( rec->serv_rp[ i ].serv_at == STATUS_OK &&
			!( rec->serv_rp[ i ].mark &
				( REC_CREATE | REC_REMOVE ) ) &&
			majority_data_one( rec, i ) < h )
		{
			rec->serv_rp[ i ].mark = REC_DATA;
			rc++;
		}
		else if ( rec->serv_ok == -1 )
		{
			rec->serv_ok = i;
		}
		i++;
	}
	if ( rc == 0 || rec->serv_ok == -1 )	goto out;

	i = 0;
	while ( rc && i < rec->serv_ln )
	{
		if ( rec->serv_rp[ i ].mark & REC_DATA )
			repair_data( rec, i );
		i++;
	}
out:
	return 0;
}

static inline int
check_directory( recovery_t * rec )
{
#if 0
fprintf( stdFp, "DIRECTORY CHECK NOW !\n" );
#endif
	return 0;
}

int
majority( recovery_t * rec )
{
	majority_type( rec );

	majority_attr( rec );

#if 1
	majority_data( rec );
#else
	if ( rec->targ_ch )
	{
		majority_data( rec );
		if ( rec->serv_ok != -1 &&
			rec->serv_rp[ serv_ok ].type == TYPE_DIR )
			check_directory( rec );
	}
#endif
}
