/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "recovery.h"

static inline void
parse_client( buffer_t * buf, recovery_t * rec )
{
	rec->xid = get_hlong( buf );
	rec->cip = get_nlong( buf );
	rec->uid = get_hlong( buf );
	rec->gid = get_hlong( buf );
#if 1
fprintf( stdFp, "RECOVERY REQUEST : CLIENT %s : ", inet_ntoa( rec->cip ) );
fprintf( stdFp, "XID %X : UID %d : GID %d\n",
		rec->xid, rec->uid, rec->gid );
#endif
}

static inline void
parse_server( buffer_t * buf, recovery_t * rec, int i )
{
	rec->serv_rp[ i ].serv_ip = get_nlong( buf );
	rec->serv_rp[ i ].serv_ex = get_hstr( buf );
#if 0
fprintf( stdFp, "TARGET SERVER %s",
		inet_ntoa( rec->serv_rp[ i ].serv_ip ) );
fprintf( stdFp, " : EXPORT PATH %s\n", rec->serv_rp[ i ].serv_ex );
#endif
}

static inline void
parse_target( buffer_t * buf, recovery_t * rec )
{
	int	i, n;

	rec->targ_ch = get_hlong( buf );
	rec->targ_pt = get_hstr( buf );

	i = 0;
	n = strlen( rec->targ_pt );
	while ( i < n && rec->targ_pt[ i ] == '/' ) i++;
	if ( i > 0 )
	{
		n -= i;
		memmove( rec->targ_pt, &rec->targ_pt[ i ], n + 1 );
	}
	i = strlen( rec->targ_pt );
	if ( i-- )
	{
		while( i >= 0 && rec->targ_pt[ i ] == '/' )
				rec->targ_pt[ i-- ] = '\0';
	}
	if ( !strlen( rec->targ_pt ) )
	{
		rec->targ_pt[ 0 ] = '.';
		rec->targ_pt[ 1 ] = '\0';
	}

	rec->targ_of = get_hlong( buf );
	rec->targ_ln = get_hlong( buf );
#if 1
fprintf( stdFp, "REQUEST TARGET %s : DATACHECK %X",
			rec->targ_pt, rec->targ_ch );
fprintf( stdFp, " : OFFSET %X : LENGTH %X\n",
			rec->targ_of, rec->targ_ln );
#endif
}

static inline unsigned long
check_target_one( int i, recovery_t * rec )
{
	unsigned long status =  STATUS_NG;

	if ( rec->serv_rp[ i ].serv_fd == -1 )
		goto out;

	memset( &rec->serv_rp[ i ].stat, 0, sizeof( struct stat ) );

	rec->sendbuff.offset = rec->sendbuff.length = 0;
	if ( !free_u32length( &rec->sendbuff, 4 ) )
		goto out;
	set_hlong( &rec->sendbuff, COMMAND_STAT );
#if 1
	set_hlong( &rec->sendbuff, 1 );
#else
	set_hlong( &rec->sendbuff, rec->targ_ch );
#endif
	set_hlong( &rec->sendbuff, rec->targ_of );
	set_hlong( &rec->sendbuff, rec->targ_ln );
	if ( !set_hstr( &rec->sendbuff, rec->targ_pt ) )
		goto out;
	if ( send_buffer( rec->serv_rp[ i ].serv_fd, &rec->sendbuff ) < 0 )
		goto out;

	rec->recvbuff.offset = 0;
	rec->recvbuff.length = to_bytelen( 6 );
	if ( receive_buffer( rec->serv_rp[ i ].serv_fd,
				&rec->recvbuff ) < to_bytelen( 6 ) )
		goto out;

	rec->serv_rp[ i ].type = get_hlong( &rec->recvbuff );
	rec->serv_rp[ i ].stat.st_mode = get_hlong( &rec->recvbuff );
	rec->serv_rp[ i ].stat.st_uid = get_hlong( &rec->recvbuff );
	rec->serv_rp[ i ].stat.st_gid = get_hlong( &rec->recvbuff );
	rec->serv_rp[ i ].stat.st_size = get_hlong( &rec->recvbuff );
	rec->serv_rp[ i ].csum = get_nlong( &rec->recvbuff );
	status = STATUS_OK;
out:
#if 0
fprintf( stdFp, "REQUEST TARGET :  %s : ", rec->targ_pt );
fprintf( stdFp, "TYPE %d : MODE %X : UID %d : GID %d : SIZE %u : CSUM %X\n",
		rec->serv_rp[ i ].type,
		rec->serv_rp[ i ].stat.st_mode,
		rec->serv_rp[ i ].stat.st_uid,
		rec->serv_rp[ i ].stat.st_gid,
		rec->serv_rp[ i ].stat.st_size,
		rec->serv_rp[ i ].csum );
#endif
	return status;
}

static inline int
check_target( recovery_t * rec )
{
	int	i = 0, c = 0;

	while ( i < rec->serv_ln )
		if ( rec->serv_rp[ i ].serv_at == STATUS_OK &&
		     check_target_one( i++, rec ) == STATUS_OK )
			c++;
	return c;
}

static inline int
target_loop( buffer_t * buf, recovery_t * rec )
{
	int	i, n;

	n = get_hlong( buf );
	for ( i = 0; i < n ; i++ )
	{
		parse_target( buf, rec );
		if ( check_target( rec ) == rec->serv_ln )
			majority( rec );
	}
	return	0;
}

static inline unsigned long
client_auth( recovery_t * rec, int i )
{
	unsigned long status =  STATUS_NG;

	rec->sendbuff.offset = rec->sendbuff.length = 0;
	if ( !free_u32length( &rec->sendbuff, 3 ) )
		goto out;
	set_hlong( &rec->sendbuff, COMMAND_AUTH );
	set_nlong( &rec->sendbuff, rec->cip );
	set_nlong( &rec->sendbuff, rec->serv_rp[ i ].serv_ip );
	if ( !set_hstr( &rec->sendbuff, rec->serv_rp[ i ].serv_ex ) )
		goto out;
	if ( send_buffer( rec->serv_rp[ i ].serv_fd, &rec->sendbuff ) < 0 )
		goto out;
	if ( receive_status( rec->serv_rp[ i ].serv_fd, &status ) < 0 )
		goto out;
out:
	return ( rec->serv_rp[ i ].serv_at = status );
}

static inline void
server_disconnect_one( recovery_t * rec, int i )
{
	if ( rec->serv_rp[ i ].serv_fd != -1 )
	{
		unsigned long status = COMMAND_EXIT;
		send_status( rec->serv_rp[ i ].serv_fd, status );
		receive_status( rec->serv_rp[ i ].serv_fd, &status );
#if 0
fprintf( stdFp, "check server connection close( %lu ).\n", status );
#endif
		close ( rec->serv_rp[ i ].serv_fd );
		rec->serv_rp[ i ].serv_fd = -1;
	}
}

static inline void
server_disconnect( recovery_t * rec )
{
	int	i = 0;

	while( i < rec->serv_ln )
		server_disconnect_one( rec, i++ );
	rec->serv_ln = rec->serv_cn = 0;
}

static inline int
server_connect_one( recovery_t * rec, int i )
{
	int	r;

	rec->serv_rp[ i ].serv_sk.sin_family = PF_INET;
	rec->serv_rp[ i ].serv_sk.sin_addr.s_addr =
				rec->serv_rp[ i ].serv_ip;
#if 0
	rec->serv_rp[ i ].serv_sk.sin_port = TARGET_CHECK_PORT;
#else
	rec->serv_rp[ i ].serv_sk.sin_port = htons( check_port );
#endif

#if 1
	rec->serv_rp[ i ].serv_fd =
		tcp_connect( &rec->serv_rp[ i ].serv_sk );
#else
	for( r = 0; r < 5; r++ )
	{
		rec->serv_rp[ i ].serv_fd =
			tcp_connect( &rec->serv_rp[ i ].serv_sk );
		if ( rec->serv_rp[ i ].serv_fd == -1 )
			sleep( 1 );
		else break;
	}
#endif
	if ( rec->serv_rp[ i ].serv_fd == -1 )
	{
#if 1
fprintf( errFp, "CHECK SERVER CONNECTION FAILED. XID ( %x ) / SERVER %s\n",
		rec->xid, inet_ntoa( rec->serv_rp[ i ].serv_ip ) );
#endif
	}

	return rec->serv_rp[ i ].serv_fd;
}

static inline int
server_connect( buffer_t * buf, recovery_t * rec )
{
	int	i;

	parse_client( buf, rec );

	if ( ( rec->serv_ln = get_hlong( buf ) ) > SERVER_MAX )
		return -1;
	for ( i = 0; i < rec->serv_ln; i++ )
	{
		parse_server( buf, rec, i );
		if ( server_connect_one( rec, i ) != -1 )
		{
			if ( !client_auth( rec, i ) )
				rec->serv_cn ++;
			else break;
		}
		else
		{
			break;
		}
	}
	return rec->serv_cn;
}

void
recovery_main( int qid )
{
	buffer_t	buf;
	recovery_t	rec;

	rec.sendbuff.buffer = rec.sbuff;
	rec.recvbuff.buffer = rec.rbuff;
	do
	{

		if ( next_pool( &buf.buffer, &buf.length ) )
		{
			int	i = 0, n = 30;

			buf.offset = 0;
			get_hlong( &buf );

			rec.serv_ln = rec.serv_cn = 0;
			if ( server_connect( &buf, &rec ) == rec.serv_ln )
			{
				fprintf( stdFp, "CLIENT AUTH OK.\n" );
				target_loop( &buf, &rec );
			}
			else
			{
				fprintf( errFp, "client auth failed.\n" );
			}

			buf.offset = 0;
			server_disconnect( &rec );
			release_pool( buf.buffer );
		}
		else
		{
			sleep( 1 );
		}
	}
	while( !sig_num );
}
