/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#ifndef __UDP_SOCKET_INCLUDE__
#define __UDP_SOCKET_INCLUDE__

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<errno.h>
#include <netdb.h>
#include<sys/types.h>
#include<sys/time.h>
#include<sys/socket.h>
#include<netinet/in.h>

struct udp_packet
{
	struct sockaddr_in	addr;
	size_t			len;
	unsigned char *		buf;
};

struct udp_socket
{
	int			fd;
	fd_set			sock_fds;
	struct sockaddr_in	addr;
	struct udp_packet	send;
	struct udp_packet	recv;
};

static inline int
udp_connect( struct udp_socket *sock )
{
	int	st;

	if ( ( sock->fd =
		socket( sock->addr.sin_family, SOCK_DGRAM, 0 ) ) != -1 )
	{
		if ( ( st = bind( sock->fd,
				(struct sockaddr *)&(sock->addr),
					sizeof(sock->addr) ) ) != -1 )
		{
			FD_ZERO( &(sock->sock_fds) );
			FD_SET( sock->fd, &(sock->sock_fds) );
		}
		else
		{
                        close( sock->fd );
                        sock->fd = st;
		}
	}
        return ((sock->fd==-1)?errno:0);
}

static inline int
udp_disconnect( struct udp_socket *sock )
{
        int	st;

	st = close( sock->fd );
	sock->fd = -1;
	FD_ZERO( &(sock->sock_fds) );

        return ((st==-1)?errno:0);
}

static inline int
udp_cansend( struct udp_socket * sock, struct timeval *timeout )
{
	int	st;
	fd_set	test_fds = sock->sock_fds;

	st = select( FD_SETSIZE, (fd_set *)NULL, &test_fds,
					(fd_set *)NULL, timeout );
        return ((st==-1)?0:1);
}

static inline int
udp_send( struct udp_socket * sock )
{
        int	st;

        st = sendto( sock->fd, sock->send.buf, sock->send.len, 0,
                        (struct sockaddr *)&(sock->send.addr),
						sizeof(sock->send.addr) );
        return ((st==-1)?errno:0);
}

static inline int
udp_canreceive( struct udp_socket * sock, struct timeval *timeout )
{
	int	st;
	fd_set	test_fds = sock->sock_fds;

	st = select( FD_SETSIZE, &test_fds, (fd_set *)NULL,
					(fd_set *)NULL, timeout );
        return ((st==-1)?0:1);
}

static inline int
udp_receive( struct udp_socket * sock )
{
        int	st, sl = sizeof(sock->addr);

	memcpy( &(sock->recv.addr), &(sock->addr), sl );
        st = recvfrom( sock->fd, sock->recv.buf, sock->recv.len, 0,
                        (struct sockaddr *)&(sock->recv.addr), &sl );
        sock->recv.len = ((st==-1)?0:st);
        return ((st==-1)?errno:0);
}

#endif
